"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var lodash_1 = require("lodash");
var number_to_words_1 = require("number-to-words");
var path_1 = require("path");
var operators_1 = require("rxjs/operators");
var transliteration_1 = require("./../utils/transliteration");
var rx_dir_1 = require("./rx.dir");
var COMPONENT_SUFFIX = 'Component';
function _getStyleExtension(aOptions) {
    return aOptions.scss ? '.scss' : '.css';
}
function _camelCase(aValue) {
    return transliteration_1.camelCase(aValue);
}
function _kebabCase(aValue) {
    return transliteration_1.kebabCase(aValue);
}
var LEADING_DIGIT = /^(\d*)/g;
var LAYOUT_SUFFIX = 'Layout';
var MODULE_SUFFIX = 'Module';
var LAYOUTS_PATH = 'layouts';
var MODULES_PATH = 'modules';
function _createLayoutComponentName(aName) {
    // trim
    var trimmedName = aName.trim();
    var fixed = trimmedName.replace(LEADING_DIGIT, function (m, p) { return (p && (p.length > 0)) ? number_to_words_1.toWords(p) + ' ' : ''; });
    /** check if the layout component starts with a number, in this case
     * convert the number to a string */
    var name = lodash_1.upperFirst(_camelCase(fixed));
    return name.endsWith(LAYOUT_SUFFIX) ? name : name + LAYOUT_SUFFIX;
}
function _createLayoutModuleName(aName) {
    // derive from layout name
    var layoutName = _createLayoutComponentName(aName);
    // replace the word 'layout'
    return layoutName.substr(0, layoutName.length - LAYOUT_SUFFIX.length) + MODULE_SUFFIX;
}
function _getPathForLayoutModule(aName, aType, aOptions) {
    // name of the layout folder segment
    var moduleName = _createLayoutModuleName(aName);
    if (moduleName.endsWith(MODULE_SUFFIX)) {
        moduleName = moduleName.substr(0, moduleName.length - MODULE_SUFFIX.length);
    }
    // layout folder
    var moduleFolder = _kebabCase(moduleName);
    // access the type properties
    var tags = aType.tags || [];
    // convert tags into segments
    var segments = aOptions.flat ? [] : tags.map(_kebabCase).sort();
    segments.push(moduleFolder);
    // return this path
    return segments.join(path_1.sep);
}
function _getPathForLayout(aName, aType, aOptions) {
    // name of the layout folder segment
    var layoutName = _createLayoutComponentName(aName);
    if (layoutName.endsWith(LAYOUT_SUFFIX)) {
        layoutName = layoutName.substr(0, layoutName.length - LAYOUT_SUFFIX.length);
    }
    // layout folder
    var layoutFolder = _kebabCase(layoutName);
    // access the type properties
    var tags = aType.tags || [];
    // convert tags into segments
    var segments = aOptions.flat ? [] : tags.map(_kebabCase).sort();
    segments.push(layoutFolder);
    // return this path
    return segments.join(path_1.sep);
}
function _getLayoutModulePath(aLayout, aType, aOptions) {
    // join
    return path_1.join(MODULES_PATH, _getPathForLayoutModule(aLayout.name, aType, aOptions));
}
function _getLayoutPath(aLayout, aType, aOptions) {
    // join
    return path_1.join(LAYOUTS_PATH, _getPathForLayout(aLayout.name, aType, aOptions));
}
/**
 * Constructs the relevant information for a components class
 *
 * @param aDir      base directory
 * @param aType     type object
 * @param aOptions  options
 * @param aLogger   logger
 *
 * @return the component object
 */
function _initLayoutClass(aDir, aCmpClass, aLayout, aMapping, aOptions, aLogger) {
    // access the type object
    var type = aCmpClass.typeDef.type;
    // check for the path suffix
    var layoutPath = _getLayoutPath(aLayout, type, aOptions);
    var layout = aLayout;
    var mapping = aMapping;
    var typeName = type.name;
    var layoutName = layout.name || typeName;
    var component = aCmpClass;
    // component names
    var layoutBaseName = _createLayoutComponentName(layoutName);
    var layoutClass = layoutBaseName + COMPONENT_SUFFIX;
    var moduleClass = _createLayoutModuleName(layoutName);
    // references
    var layoutBaseRef = _camelCase(layoutBaseName);
    var layoutRef = layoutBaseRef;
    var moduleRef = _camelCase(moduleClass);
    // file names
    var layoutFile = layoutRef + '.ts';
    var moduleFile = moduleRef + '.ts';
    // template and style
    var layoutTemplateFile = layoutBaseRef + '.html';
    var layoutStyleFile = layoutBaseRef + _getStyleExtension(aOptions);
    // selector
    var layoutSelector = _kebabCase('app ' + layoutClass);
    var layoutTemplate = layout.template;
    // the base path for the layout
    var folder = path_1.normalize(path_1.join(aDir, layoutPath));
    var result = {
        component: component,
        type: type, mapping: mapping, layout: layout, folder: folder,
        layoutClass: layoutClass, moduleClass: moduleClass,
        layoutRef: layoutRef, moduleRef: moduleRef,
        layoutFile: layoutFile, moduleFile: moduleFile,
        layoutTemplateFile: layoutTemplateFile, layoutStyleFile: layoutStyleFile,
        layoutSelector: layoutSelector, layoutTemplate: layoutTemplate
    };
    // log this
    aLogger.next("Type: " + typeName + ", LayoutClass: " + JSON.stringify(layoutClass) + ", Folder: " + folder);
    // ok
    return result;
}
/**
 * Finds the directory tha the component should be created in. Tests
 * each potential source directory for an existing component file. If non exists, uses the
 * first source directory
 *
 * @param aType        type
 * @param aSources     sources directory
 *
 * @return the selected source directory
 */
function _getLayoutsDir(aLayout, aMapping, aType, aSources, aOptions, aMkdirp, aLogger) {
    // check for the path suffix
    var typePath = _getLayoutPath(aLayout, aType, aOptions);
    var className = _createLayoutComponentName(aLayout.name);
    var relPath = path_1.join(typePath, _camelCase(className) + '.ts');
    // map each source to a check
    return rx_dir_1.rxFindDir(aSources, relPath, aMkdirp).pipe(operators_1.tap(function (dir) { return aLogger.next("Selected layout source directory [" + dir + "] for layout [" + className + "]."); }));
}
function _createLayoutClass(aLayout, aMapping, aSources, aCmpClass, aOptions, aMkdirp, aLogger) {
    // the type
    var type = aCmpClass.typeDef.type;
    // locate the component directory
    return _getLayoutsDir(aLayout, aMapping, type, aSources, aOptions, aMkdirp, aLogger).pipe(operators_1.first(), operators_1.map(function (dir) { return _initLayoutClass(dir, aCmpClass, aLayout, aMapping, aOptions, aLogger); }), operators_1.mergeMap(function (clz) { return aMkdirp(clz.folder).pipe(operators_1.mapTo(clz)); }), operators_1.shareReplay());
}
var LayoutRegistry = /** @class */ (function () {
    function LayoutRegistry(aBaseFolders, aOptions, aMkdirp, aComponentRegistry, aLogger) {
        this.aBaseFolders = aBaseFolders;
        this.aOptions = aOptions;
        this.aMkdirp = aMkdirp;
        this.aComponentRegistry = aComponentRegistry;
        this.aLogger = aLogger;
        this.layoutMap = {
            byId: {},
            byName: {},
            dirById: {}
        };
        this.layoutRegistry = {};
    }
    LayoutRegistry.prototype.registerLayout = function (aLayout) {
        var layout = aLayout;
        var lid = layout.id;
        var lname = layout.name;
        if (this.layoutMap.byId[lid] || this.layoutMap.byName[lname]) {
            // do not create a layout, twice
            return false;
        }
        // register
        this.layoutMap.byId[lid] = layout;
        this.layoutMap.byName[lname] = layout;
        // ok
        return true;
    };
    LayoutRegistry.prototype.findLayoutClass = function (aLayout, aMapping, aType) {
        var _this = this;
        // the name
        var layoutId = aLayout.id;
        // check if we are already searching
        var rxRunning = this.layoutRegistry[layoutId];
        if (!rxRunning) {
            // find the component
            var rxComponent = this.aComponentRegistry.findComponentClass(aType);
            // try to find the folder
            rxRunning = this.layoutRegistry[layoutId] =
                rxComponent.pipe(operators_1.mergeMap(function (cmp) { return _createLayoutClass(aLayout, aMapping, _this.aBaseFolders, cmp, _this.aOptions, _this.aMkdirp, _this.aLogger); }));
        }
        // ok
        return rxRunning;
    };
    return LayoutRegistry;
}());
exports.LayoutRegistry = LayoutRegistry;
