/* Copyright IBM Corp. 2017 */
import { escape as escapeHTML } from 'lodash';
import { join } from 'path';
import { Observable, of, Subject } from 'rxjs';
import { map, mergeMap } from 'rxjs/operators';
import { parse as parseURL } from 'url';

import { firstOf } from '../utils/file.utils';
import { rxForkJoin } from '../utils/rx.utils';
import {
    CMD_ADD,
    OPTION_DATA,
    OPTION_NAME,
    OPTION_RELATIVE,
    OPTION_ROOT,
    OPTION_SOURCE,
    OPTION_URL,
    OPTION_VERBOSE,
} from './../constants';
import { Options, optNormalizeDir } from './../options';
import { configIndexDirFromContextRoot, configTrimSlashes } from './../utils/config';
import { readFile, writeFile } from './../utils/rx.walk';
import { ensureTrailingSlash } from './../utils/url.utils';
import { wchDeliveryUrlFromApiUrl } from './../utils/wchtools';
import { AbstractCommand } from './abstract.command';
import { Command } from './command';
import {
    HELP_API_URL,
    HELP_APPLICATION_NAME,
    HELP_CONTEXT_ROOT,
    HELP_DATA_DIRECTORY,
    HELP_INDEX_SOURCE,
    HELP_RELATIVE,
    HELP_VERBOSE,
} from './help';

function _createUrl(aContextRoot: string, aApiBaseUrl: string, aRelative: boolean): Observable<string> {
    // the delivery URL
    const absWchDeliveryUrl = ensureTrailingSlash(wchDeliveryUrlFromApiUrl(aApiBaseUrl));
    // path of the index file
    const absRootUrl = ensureTrailingSlash(`${absWchDeliveryUrl}${configTrimSlashes(aContextRoot)}`);
    // parse the URL
    if (!aRelative) {
        return of(absRootUrl);
    }
    // make relative
    const parsed = parseURL(absRootUrl);
    return of(parsed.path || '/');
}

function _fixIndex(aSource: string, aUrl: string, aLogger: Subject<string>): string {
    // log the settings
    aLogger.next(`BaseURL [${aUrl}].`);
    // insert the layouts import
    let src = aSource;
    // add the module to the end of the declarations
    src = src.replace(/(<base\s+href=\")([^\"]*)(\">)/g, '$1' + escapeHTML(aUrl) + '$3');

    return src;
}

function _rewriteIndex(aSrcDir: string,
    aContextRoot: string,
    aApiBaseUrl: string,
    aRelative: boolean,
    aLogger: Subject<string>): Observable<string> {
    // compute the proper base URL
    const baseUrl = _createUrl(aContextRoot, aApiBaseUrl, aRelative);
    // name of the index file
    const name = join(aSrcDir, 'index.html');
    return readFile(name).pipe(
        mergeMap(src => baseUrl.pipe(map(url => _fixIndex(src, url, aLogger)))),
        mergeMap(data => writeFile(name, data))
    );
}

/**
 * Generates layout and layout mappings
 *
 * add base --data DATA_FOLDER --src SRC_FOLDER [--relative]
 */
export class AddBaseCommand extends AbstractCommand implements Command {

    static readonly TYPE = 'base';

    static readonly COMMAND = CMD_ADD;

    constructor(options: Options) {
        super(options);
    }

    printHelp(): void {
        // log the command line
        console.log(AddBaseCommand.COMMAND, AddBaseCommand.TYPE,
            `[--${OPTION_SOURCE} <SOURCE>]`,
            `[--${OPTION_DATA} <DATA>]`,
            `[--${OPTION_URL} <URL>]`,
            `[--${OPTION_NAME} <NAME>]`,
            `[--${OPTION_ROOT} <ROOT>]`,
            `[--${OPTION_RELATIVE}]`,
            `[--${OPTION_VERBOSE}]`);
        // help on the options
        console.log(`--${OPTION_SOURCE} <SOURCE>`, ':', HELP_INDEX_SOURCE);
        console.log(`--${OPTION_DATA} <DATA>`, ':', HELP_DATA_DIRECTORY);
        console.log(`--${OPTION_URL} <URL>`, ':', HELP_API_URL);
        console.log(`--${OPTION_NAME} <NAME>`, ':', HELP_APPLICATION_NAME);
        console.log(`--${OPTION_ROOT}`, ':', HELP_CONTEXT_ROOT);
        console.log(`--${OPTION_VERBOSE}`, ':', HELP_VERBOSE);
        console.log(`--${OPTION_RELATIVE}`, ':', HELP_RELATIVE);
    }

    process(): Observable<string> {
        // options
        const opts = this.getOptions();
        const dataDirOpt = this.getDataDir();
        const appName = opts.name || this.getAppName();
        const contextRoot = `/${configTrimSlashes(opts.root || this.getContextRoot(appName))}`;
        // the directories
        const srcDirOpt = optNormalizeDir(firstOf(opts.src) || configIndexDirFromContextRoot(contextRoot, dataDirOpt));
        const srcDir = this.rxMkdirp(srcDirOpt);
        // base URL
        const apiBaseUrl = this.getApiBase();
        // check if we should use a relative URL
        const bRelative = !!opts.relative;

        // the logger
        const logger = this.getLogger();

        // log the settings
        logger.next(`Source dir [${srcDirOpt}].`);

        // wait for the directories
        return rxForkJoin([srcDir, apiBaseUrl]).pipe(
            mergeMap(([aSrcDir, aBaseUrl]) => _rewriteIndex(aSrcDir, contextRoot, aBaseUrl, bRelative, logger))
        );
    }
}
