/* Copyright IBM Corp. 2017 */
import { join } from 'path';
import { combineLatest, Observable } from 'rxjs';
import { map, mergeMap } from 'rxjs/operators';

import { firstOf } from '../utils/file.utils';
import { CMD_ADD, OPTION_DATA, OPTION_NAME, OPTION_ROOT, OPTION_SOURCE, OPTION_VERBOSE } from './../constants';
import { Options, optNormalizeDir } from './../options';
import { configIndexDirFromContextRoot, configTrimSlashes } from './../utils/config';
import { hbsCompile } from './../utils/rx.hbs';
import { readFile as rxReadFile, writeFile as rxWriteFile } from './../utils/rx.walk';
import { AbstractCommand } from './abstract.command';
import { Command } from './command';
import { HELP_APPLICATION_NAME, HELP_CONTEXT_ROOT, HELP_DATA_DIRECTORY, HELP_INDEX_SOURCE, HELP_VERBOSE } from './help';

function _getPolyfillTemplate(): Observable<HandlebarsTemplateDelegate> {
    return hbsCompile(join(__dirname, '..', '..', '..', 'assets', 'esi/esi.hbs'));
}

const HEAD_REGEXP = /<\/head>/g;

function _insertMarkup(aPolyfill: string, aMarkup: string): string {
    // replace the head regular expression
    return aMarkup.replace(HEAD_REGEXP, `${aPolyfill}\n</head>`);
}

function _getPolyfillScript(): Observable<string> {
    return rxReadFile(join(__dirname, '..', '..', '..', 'assets', 'esi/esi.min.js'));
}

function _rewriteIndex(aSrcDir: string): Observable<string> {
    // load the polyfill script as context
    const rxPolyfillContext = _getPolyfillScript().pipe(
        map(script => ({
            SCRIPT: script
        })));
    // load the markup template
    const rxPolyfillTemplate = _getPolyfillTemplate();
    // excute
    const rxMarkup = combineLatest(rxPolyfillContext, rxPolyfillTemplate).pipe(
        map(([ctx, tmp]) => tmp(ctx)));

    // read the index file
    const fileName = join(aSrcDir, 'index.html');
    return combineLatest(rxReadFile(fileName), rxMarkup).pipe(
        map(([index, markup]) => _insertMarkup(markup, index)),
        mergeMap(newMarkup => rxWriteFile(fileName, newMarkup))
    );
}

export class AddEsiPolyfillCommand extends AbstractCommand implements Command {

    static readonly TYPE = 'esi-polyfill';

    static readonly COMMAND = CMD_ADD;

    constructor(private options: Options) {
        super(options);
    }

    printHelp(): void {
        // log the command line
        console.log(AddEsiPolyfillCommand.COMMAND, AddEsiPolyfillCommand.TYPE,
            `[--${OPTION_SOURCE} <SOURCE>]`,
            `[--${OPTION_DATA} <DATA>]`,
            `[--${OPTION_NAME} <NAME>]`,
            `[--${OPTION_ROOT} <ROOT>]`,
            `[--${OPTION_VERBOSE}]`);
        // help on the options
        console.log(`--${OPTION_SOURCE} <SOURCE>`, ':', HELP_INDEX_SOURCE);
        console.log(`--${OPTION_DATA} <DATA>`, ':', HELP_DATA_DIRECTORY);
        console.log(`--${OPTION_NAME} <NAME>`, ':', HELP_APPLICATION_NAME);
        console.log(`--${OPTION_ROOT}`, ':', HELP_CONTEXT_ROOT);
        console.log(`--${OPTION_VERBOSE}`, ':', HELP_VERBOSE);
    }

    process(): Observable<string> {
        // logger
        const logger = this.getLogger();
        // data directory
        const opts = this.options;
        // the directories
        const dataDirOpt = this.getDataDir();
        const appName = opts.name || this.getAppName();
        const contextRoot = `/${configTrimSlashes(opts.root || this.getContextRoot(appName))}`;
        // the directories
        const srcDirOpt = optNormalizeDir(firstOf(opts.src) || configIndexDirFromContextRoot(contextRoot, dataDirOpt));
        const rxSrcDir = this.rxMkdirp(srcDirOpt);
        // rewrite
        return rxSrcDir.pipe(mergeMap(_rewriteIndex));
    }

}
