/* Copyright IBM Corp. 2017 */
import * as Promise from 'bluebird';
import { join } from 'path';
import * as RelateUrl from 'relateurl';
import { Observable, Observer, Subject } from 'rxjs';
import { mergeMap } from 'rxjs/operators';
import { parse as parseURL } from 'url';

import { firstOf } from '../utils/file.utils';
import {
    CMD_BUILD,
    OPTION_DATA,
    OPTION_NAME,
    OPTION_PASSWORD,
    OPTION_RELATIVE,
    OPTION_SOURCE,
    OPTION_URL,
    OPTION_USER,
    OPTION_VERBOSE,
} from './../constants';
import { Options, optNormalizeDir } from './../options';
import { configGetBaseUrl, configTrimSlashes } from './../utils/config';
import { ensureTrailingSlash } from './../utils/url.utils';
import { wchDeliveryUrlFromApiUrl } from './../utils/wchtools';
import { AbstractCommand } from './abstract.command';
import { Command } from './command';
import {
    HELP_API_URL,
    HELP_APPLICATION_NAME,
    HELP_BUILD_SOURCE,
    HELP_DATA_DIRECTORY,
    HELP_PASSWORD,
    HELP_RELATIVE,
    HELP_USERNAME,
    HELP_VERBOSE,
} from './help';

const cli = require('@angular/cli').default;
function _buildDevelopment(
    aAppName: string,
    aContextRoot: string,
    aApiBaseUrl: string,
    aSrcDir: string,
    aRelative: boolean,
    aLogger: Subject<string>): Observable<string> {
    // the result
    return Observable.create((files: Observer<string>) => {
        // the delivery URL
        const absWchDeliveryUrl = ensureTrailingSlash(wchDeliveryUrlFromApiUrl(aApiBaseUrl));
        // path of the index file
        const absRootUrl = ensureTrailingSlash(`${absWchDeliveryUrl}${configTrimSlashes(aContextRoot)}`);
        // get the base URL
        const absAppUrl = ensureTrailingSlash(configGetBaseUrl(aAppName, aApiBaseUrl));
        // dist directory
        const distDir = join(aSrcDir, 'dist');
        // the base href
        let baseUrl;
        // check if we want a relative URL
        if (aRelative) {
            // parse
            const parsedUrl = parseURL(absRootUrl);
            baseUrl = parsedUrl.path || '/';
        } else {
            // use the absolute URL
            baseUrl = absRootUrl;
        }
        // the relative delivery URL
        const resolve = new RelateUrl(absRootUrl);
        const deliveryUrl = resolve.relate(absAppUrl);
        // log this
        aLogger.next(`BaseUrl [${baseUrl}], deliveryUrl [${deliveryUrl}], source directory [${aSrcDir}].`);
        // command line interface
        const args = ['build', '--base-href', baseUrl];
        if (deliveryUrl && (deliveryUrl.length > 0)) {
            args.push('--deploy-url', deliveryUrl);
        }
        aLogger.next(`ng command line arguments ${JSON.stringify(args)}.`);
        // execute the command, we will sync on files
        Promise.resolve(cli({
            cliArgs: args,
            directory: aSrcDir,
            inputStream: process.stdin,
            outputStream: process.stdout
        }))
            .then(() => files.complete())
            .catch(err => files.error(err));
    });
}

export class BuildDevelopmentCommand extends AbstractCommand implements Command {

    static readonly TYPE = 'development';
    static readonly COMMAND = CMD_BUILD;

    constructor(options: Options) {
        super(options);
    }

    printHelp(): void {
        // log the command line
        console.log(BuildDevelopmentCommand.COMMAND, BuildDevelopmentCommand.TYPE,
            `[--${OPTION_SOURCE} <SOURCE>]`,
            `[--${OPTION_NAME} <NAME>]`,
            `[--${OPTION_DATA} <DATA>]`,
            `[--${OPTION_URL} <URL>]`,
            `[--${OPTION_USER} <USER>]`,
            `[--${OPTION_PASSWORD} <PASSWORD>]`,
            `[--${OPTION_VERBOSE}]`);
        // help on the options
        console.log(`--${OPTION_SOURCE} <SOURCE>`, ':', HELP_BUILD_SOURCE);
        console.log(`--${OPTION_NAME} <NAME>`, ':', HELP_APPLICATION_NAME);
        console.log(`--${OPTION_DATA} <DATA>`, ':', HELP_DATA_DIRECTORY);
        console.log(`--${OPTION_URL} <URL>`, ':', HELP_API_URL);
        console.log(`--${OPTION_USER} <USER>`, ':', HELP_USERNAME);
        console.log(`--${OPTION_PASSWORD} <PASSWORD>`, ':', HELP_PASSWORD);
        console.log(`--${OPTION_VERBOSE}`, ':', HELP_VERBOSE);
        console.log(`--${OPTION_RELATIVE}`, ':', HELP_RELATIVE);
    }

    process(): Observable<string> {
        // the options
        const opts = this.getOptions();
        // get the tools
        const apiBaseUrl = this.getApiBase();
        const appName = opts.name || this.getAppName();
        const contextRoot = `/${configTrimSlashes(opts.root || this.getContextRoot(appName))}`;
        // source directory
        const srcDir = optNormalizeDir(firstOf(opts.src) || this.getPackageDir());
        // relative flag
        const bRelative = !!opts.relative;
        // logger
        const logger = this.getLogger();
        // log the config
        logger.next(`AppName [${appName}], contextRoot [${contextRoot}]`);
        // process
        return apiBaseUrl.pipe(mergeMap(aApiBaseUrl => _buildDevelopment(appName, contextRoot, aApiBaseUrl, srcDir, bRelative, logger)));
    }
}
