/* Copyright IBM Corp. 2017 */
import { join } from 'path';
import { concat, Observable, Subject } from 'rxjs';
import { distinct, mergeMap } from 'rxjs/operators';

import { rxCopyDependencies } from '../utils/asset.dependencies';
import { writeJsonFileSafe } from '../utils/rx.file';
import { rxMkdirp } from '../utils/rx.walk';
import {
    CMD_COPY,
    OPTION_AGGREGATED,
    OPTION_DATA,
    OPTION_OUT,
    OPTION_OVERRIDE,
    OPTION_PASSWORD,
    OPTION_PKG_DIR,
    OPTION_URL,
    OPTION_USER,
    OPTION_VERBOSE,
} from './../constants';
import { Options, optNormalizeDir } from './../options';
import { rxCopyDir } from './../utils/rx.copy';
import { rxDeleteDir } from './../utils/rx.delete';
import { FILE_WCHTOOLS_JSON, KEY_API_URL, wchToolsGetOptions, WchToolsOptions } from './../utils/wchtools';
import { AbstractCommand } from './abstract.command';
import { Command } from './command';
import {
    HELP_AGGREGATED,
    HELP_API_URL,
    HELP_DATA_DIRECTORY,
    HELP_DST_DIR,
    HELP_OVERRIDE,
    HELP_PASSWORD,
    HELP_PKG_DIR,
    HELP_USERNAME,
    HELP_VERBOSE,
} from './help';

function _copyAggregated(aPackageFile: string, aDstDir: string, aLogger: Subject<string>): Observable<string> {
    const pkgFile = aPackageFile;
    // copy the dependencies
    const dstBase = aDstDir;

    aLogger.next(`Aggregating to [${dstBase}].`);

    // clean up after execution
    const rxDelete = rxDeleteDir(dstBase);

    // copy the dependencies
    const rxCopy = rxCopyDependencies(pkgFile, dstBase);

    return rxCopy;
}

function _copyDirect(aSrcDir: string, aDstDir: string, aLogger: Subject<string>): Observable<string> {
    // log this
    aLogger.next(`Copying from [${aSrcDir}] to [${aDstDir}].`);
    // copy source files
    return rxCopyDir(aSrcDir, aDstDir);
}

function _writeOptions(aTools: WchToolsOptions, aDstDir: string): Observable<string> {
    // prepare the context
    const data = {
        [KEY_API_URL]: aTools.baseUrl
    };
    // write the file
    return rxMkdirp(aDstDir).pipe(
        mergeMap(dir => writeJsonFileSafe(join(dir, FILE_WCHTOOLS_JSON), data, true))
    );
}

/**
 * Copies the build output to the correct location
*/
export class CopyDataCommand extends AbstractCommand implements Command {

    static readonly TYPE = 'data';
    static readonly COMMAND = CMD_COPY;

    constructor(options: Options) {
        super(options);
    }

    printHelp(): void {
        // log the command line
        console.log(CopyDataCommand.COMMAND, CopyDataCommand.TYPE,
            `--${OPTION_PKG_DIR} <PKG>`,
            `--${OPTION_AGGREGATED}`,
            `--${OPTION_DATA} <DATA>`,
            `--${OPTION_OUT} <OUT>`,
            `[--${OPTION_USER} <USER>]`,
            `[--${OPTION_PASSWORD} <PWD>]`,
            `[--${OPTION_URL} <URL>]`,
            `[--${OPTION_OVERRIDE}]`,
            `[--${OPTION_VERBOSE}]`);
        // log the parameters
        console.log(`--${OPTION_PKG_DIR} <PKG>`, ':', HELP_PKG_DIR);
        console.log(`--${OPTION_AGGREGATED}`, ':', HELP_AGGREGATED);
        console.log(`--${OPTION_OUT} <OUT>`, ':', HELP_DST_DIR);
        console.log(`--${OPTION_DATA} <DATA>`, ':', HELP_DATA_DIRECTORY);
        console.log(`--${OPTION_USER} <USER>`, ':', HELP_USERNAME);
        console.log(`--${OPTION_PASSWORD} <PWD>`, ':', HELP_PASSWORD);
        console.log(`--${OPTION_URL} <URL>`, ':', HELP_API_URL);
        console.log(`--${OPTION_OVERRIDE}`, ':', HELP_OVERRIDE);
        console.log(`--${OPTION_VERBOSE}`, ':', HELP_VERBOSE);
    }

    process(): Observable<string> {
        // the options
        const opts = this.getOptions();
        // logger
        const logger = this.getLogger();
        // check if we are aggregated
        const bAggregated = !!opts.aggregated;
        // get the output directory
        const dstDir = optNormalizeDir(opts.out || 'tmp');
        // depending on this, perform the copy
        const rxCopiedFiles = bAggregated ? _copyAggregated(join(this.getPackageDir(), 'package.json'), dstDir, logger) : _copyDirect(this.getDataDir(), dstDir, logger);
        // wchtools options
        const rxTools = wchToolsGetOptions(this.getOptions(), this.getDataDir());
        const rxOptions = rxTools.pipe(mergeMap(opts => _writeOptions(opts, dstDir)));
        // concatenate
        return concat(rxCopiedFiles, rxOptions).pipe(distinct());
    }
}