/* Copyright IBM Corp. 2017 */
import { createHash } from 'crypto';
import { getType } from 'mime';
import { join, parse } from 'path';
import { merge, Observable, Subject } from 'rxjs';
import { map, mergeMap } from 'rxjs/operators';
import { v4 } from 'uuid';

import { firstOf } from '../utils/file.utils';
import { CMD_CREATE } from './../constants';
import { Options, optNormalizeDir } from './../options';
import { canonicalizeJSON } from './../utils/json';
import { addRevisionId } from './../utils/revision.utils';
import { readFile as rxReadFile, rxCopyFile, rxMkdirp, rxStats, writeFile as rxWriteFile } from './../utils/rx.walk';
import { AbstractCommand } from './abstract.command';
import { Command } from './command';

function _fixAssetType(aAsset: any, aSrc: string, aName: string): Observable<any> {
    // parse the filename
    const parserName = parse(aSrc);
    // generate the md5
    const rxMd5 = rxReadFile(aSrc).pipe(
        map(data => createHash('md5').update(data).digest('base64'))
    );

    // stats
    const stats = rxStats(aSrc);

    // resource id
    const resId = v4();

    // add ID
    aAsset.id = v4();
    aAsset.name = aName;
    aAsset.mediaType = getType(aSrc);
    // construct a path
    const damid = v4();
    const prefix = damid.substr(0, 2);
    // final path
    const path = `/dxdam/${prefix}/${damid}/${aName}`;
    aAsset.path = path;
    // type
    aAsset.assetType = aAsset.mediaType.split('/')[0];
    aAsset.isManaged = true;
    aAsset.fileName = parserName.base;
    // resource infos
    aAsset.resource = resId;
    // renditions support
    aAsset.renditions = {
        'default': {
            'id': `r=${resId}&a=${aAsset.id}`,
            'source': `/authoring/v1/resources/${resId}`
        }
    };
    // add stats
    return stats.pipe(map(s => aAsset.fileSize = s.size),
        mergeMap(() => rxMd5.pipe(map(digest => {
            aAsset.digest = digest;
            return canonicalizeJSON(addRevisionId(aAsset));
        }))));
}

function _createAssetDoc(aSrc: string, aName: string): Observable<any> {
    // load the template
    return rxReadFile(join(__dirname, '..', '..', '..', 'assets', 'asset.json')).pipe(
        map(data => JSON.parse(data)),
        mergeMap(template => _fixAssetType(template, aSrc, aName))
    );
}

function _writeAsset(aSrc: string, aName: string, aDataDir: string, aLogger: Subject<string>): Observable<string> {
    // construct the object
    return _createAssetDoc(aSrc, aName).pipe(
        mergeMap(doc => {
            // get the path
            const path = doc.path;
            const parsedPath = parse(path);
            // target dir
            const dirAsset = join(aDataDir, 'assets', parsedPath.dir);
            const dir = rxMkdirp(dirAsset);
            // copy
            const fileCopy = rxCopyFile(aSrc, join(aDataDir, 'assets', path));
            // write the file
            const filename = `${dirAsset}/${aName}_amd.json`;
            const jsonFile = dir.pipe(mergeMap(d => rxWriteFile(filename, JSON.stringify(doc, undefined, 2))));
            // merge
            return merge(fileCopy, jsonFile);
        })
    );
}

/**
 * Generates angular components based on type and layout mapping definitions
 *
 * create assert --data DATA_FOLDER --src ASSET_FILE
 */
export class CreateAssetCommand extends AbstractCommand implements Command {

    static readonly TYPE = 'asset';

    static readonly COMMAND = CMD_CREATE;

    constructor(private options: Options) {
        super(options);
    }

    process(): Observable<string> {
        // logger
        const logger = this.getLogger();
        // data directory
        const opts = this.options;
        const dataDirOpt = this.getDataDir();
        const rxDataDir = this.rxMkdirp(dataDirOpt);
        // source
        const src = optNormalizeDir(firstOf(opts.src) || 'undefined.jpg');
        // parse
        const parsedSrc = parse(src);
        const name = opts.name || parsedSrc.base;
        // log this
        logger.next(`Data dir [${dataDirOpt}], source file [${src}].`);
        // load wchtools
        return rxDataDir.pipe(mergeMap(aDataDir => _writeAsset(src, name, aDataDir, logger)));
    }
}

export {
    _writeAsset as writeAsset
};
