/* Copyright IBM Corp. 2017 */
import { cloneDeep } from 'lodash';
import { getType } from 'mime';
import { join, parse } from 'path';
import { combineLatest, merge, Observable, Subject } from 'rxjs';
import { filter, map, mergeMap, shareReplay } from 'rxjs/operators';
import { v4 } from 'uuid';

import { firstOf } from '../utils/file.utils';
import { CMD_CREATE } from './../constants';
import { Options, optNormalizeDir } from './../options';
import { canonicalizeJSON } from './../utils/json';
import { addRevisionId } from './../utils/revision.utils';
import { readFile as rxReadFile, walkFiles, writeFile as rxWriteFile } from './../utils/rx.walk';
import { WchTools } from './../utils/wchtools';
import { AbstractCommand } from './abstract.command';
import { Command } from './command';
import { writeAsset } from './create.asset.command';

/**
 * Iterates over all image files
 *
 * @param aSrcFolder    the source folder
 * @return all image files
 */
function _getImages(aSrcFolder: string): Observable<string> {
    // walk
    return walkFiles(aSrcFolder).pipe(
        filter(f => f.stats.isFile()),
        filter(f => f && (getType(f.absPath) || 'application/octet-stream').startsWith('image/')),
        map(f => f.absPath)
    );
}

function _loadTemplate(): Observable<any> {
    return rxReadFile(join(__dirname, '..', '..', '..', 'assets', 'image.content.json')).pipe(
        map(data => JSON.parse(data))
    );
}

function _createImage(aTypeId: string,
    aSrcImage: string,
    aContentTemplate: any,
    aDataDir: string,
    aLogger: Subject<string>): Observable<string> {
    // construct the asset
    const parsedName = parse(aSrcImage);
    const assetName = parsedName.base;
    // template
    const contentId = v4();
    const tmp = cloneDeep(aContentTemplate);
    tmp.id = contentId;
    tmp.name = assetName;
    tmp.typeId = aTypeId;

    // target filename
    const tgtName = join(aDataDir, 'content', `${contentId}_cmd.json`);

    const asset = writeAsset(aSrcImage, assetName, aDataDir, aLogger).pipe(shareReplay());
    // load the asset descriptor
    const assetDesc = asset.pipe(
        filter(name => name.endsWith('_amd.json')),
        mergeMap(f => rxReadFile(f)),
        map(data => JSON.parse(data)),
        shareReplay()
    );
    // get the ID
    const assetId = assetDesc.pipe(map((j: any) => j.id));
    // construct the type
    const content =
        assetDesc.pipe(
            map((desc: any) => {
                // update with some metadata
                tmp.elements.image.renditions = {
                    'default': {
                        'renditionId': desc.renditions.default.id,
                        'source': desc.renditions.default.source
                    }
                };
                tmp.elements.image.asset.resourceUri = desc.renditions.default.source;
                tmp.elements.image.asset.fileName = desc.fileName;
                tmp.elements.image.asset.fileSize = desc.fileSize;
                tmp.elements.image.asset.mediaType = desc.mediaType;
            }),
            mergeMap(() => assetId.pipe(aid => {
                // update the type
                tmp.elements.image.asset.id = aid;
                tmp.thumbnail.id = aid;
                // ok
                return canonicalizeJSON(addRevisionId(tmp));
            })),
            mergeMap(t => rxWriteFile(tgtName, JSON.stringify(t, undefined, 2)))
        );
    // merge
    const all = merge(asset, content).pipe(shareReplay());
    // log when complete
    all.subscribe(undefined, undefined, () => aLogger.next(`Image [${aSrcImage}].`));
    // ok
    return all;
}

function _findTypeId(aTypeName: string, aTools: WchTools): string {
    // locate the type
    const types = aTools.types;
    return Object.keys(types).find(id => (id === aTypeName) || (types[id].name === aTypeName)) || aTypeName;
}

function _writeImages(aTypeName: string,
    aSrcDir: string,
    aDataDir: string,
    aWchTools: WchTools,
    aLogger: Subject<string>): Observable<string> {
    // log this
    aLogger.next(`Writing images ...`);
    // locate the type id
    const typeId = _findTypeId(aTypeName, aWchTools);
    // load the template
    const tmp = _loadTemplate();
    // load the image
    return tmp.pipe(mergeMap(tmpl => _getImages(aSrcDir).pipe(
        mergeMap(img => _createImage(typeId, img, tmpl, aDataDir, aLogger)))
    ));
}

/**
 * Generates angular components based on type and layout mapping definitions
 *
 * create images --data DATA_FOLDER --type type --src SOURCE_FOLDER
 */
export class CreateImagesCommand extends AbstractCommand implements Command {

    static readonly TYPE = 'images';

    static readonly COMMAND = CMD_CREATE;

    constructor(private options: Options) {
        super(options);
    }

    process(): Observable<string> {
        // logger
        const logger = this.getLogger();
        // data directory
        // current options
        const opts = this.options;
        const dataDirOpt = this.getDataDir();
        const rxDataDir = this.rxMkdirp(dataDirOpt);
        // source
        const sourceDirOpt = optNormalizeDir(firstOf(opts.src) || '.');
        // type
        const type = opts.type || 'Photo';
        // load the tools
        const rxWchTools = this.getWchTools();
        // log this
        logger.next(`Data dir [${dataDirOpt}], source files [${sourceDirOpt}], content type [${type}].`);
        // load wchtools
        return combineLatest(rxDataDir, rxWchTools).pipe(
            mergeMap(([aDataDir, aWchTools]) => _writeImages(type, sourceDirOpt, aDataDir, aWchTools, logger))
        );
    }
}
