/* Copyright IBM Corp. 2017 */
import { readFile } from 'graceful-fs';
import { join, resolve } from 'path';
import { from, Observable } from 'rxjs';
import {
    ClassDeclaration,
    CompilerOptions,
    createProgram,
    forEachChild,
    isCallExpression,
    isClassDeclaration,
    isModuleDeclaration,
    ModuleResolutionKind,
    Node,
    Program,
    TypeChecker,
} from 'typescript';

import { firstOf } from '../utils/file.utils';
import { CMD_CREATE } from './../constants';
import { Options } from './../options';
import { DecoratorMap, tsGetDecorators } from './../utils/typescript.utils';
import { AbstractCommand } from './abstract.command';
import { Command } from './command';


const KEY_LAYOUT_COMPONENT = 'LayoutComponent';
const KEY_COMPONENT = 'Component';

interface Layout {
    className: string;
    selectors: string[];
}

function _decodeLayout(aDecs: DecoratorMap, aClass: ClassDeclaration): Layout|null {
    const layout = aDecs[KEY_COMPONENT];

    if (isCallExpression(layout.expression)) {
        // get the arguments
        const args = layout.expression.arguments;
        if (args.length > 0) {
            // first arg is the array

        }

        console.log('layout', layout.expression.arguments);
        layout.expression.arguments.forEach(a => forEachChild(a, c => console.log('child', c)));
    }


    return null;
}

function _isLayoutComponent(aDecs: DecoratorMap) {
    return aDecs[KEY_LAYOUT_COMPONENT] && aDecs[KEY_COMPONENT];
}

/** visit nodes finding classes */
function _visit(node: Node, checker: TypeChecker) {

    if (isClassDeclaration(node)) {
        const cd = node as ClassDeclaration;
        const decs = tsGetDecorators(cd);
        if (_isLayoutComponent(decs)) {
            _decodeLayout(decs, cd);
        } else {
            // This is a top level class, get its symbol
            forEachChild(node, f => _visit(f, checker));
        }
    } else
    if (isModuleDeclaration(node)) {
        // This is a namespace, visit its children
        forEachChild(node, f => _visit(f, checker));
    }
}

function _getClasses(aProgram: Program) {
    // the type checker
    const typeChecker = aProgram.getTypeChecker();
    // iterate
    // Visit every sourceFile in the program
    for (const sourceFile of aProgram.getSourceFiles()) {
        // Walk the tree to search for classes
        forEachChild(sourceFile, f => _visit(f, typeChecker));
    }
}

function _readCompilerOptions(aPath: string): Promise<CompilerOptions> {
    return new Promise((aResolve, aReject) => readFile(aPath, 'utf-8', (err, data) => {
        if (err) {
            aReject(err);
        } else {
            // parse the options
            const opt = JSON.parse(data).compilerOptions as CompilerOptions;
            opt.configFilePath = aPath;
            opt.moduleResolution = ModuleResolutionKind.NodeJs;
            aResolve(opt);
        }
    }));
}

export class CreateMappingsCommand extends AbstractCommand implements Command {

    static readonly TYPE = 'mappings';

    static readonly COMMAND = CMD_CREATE;

    constructor(private options: Options) {
        super(options);
    }

    process(): Observable<string> {
        const opt = this.options;
        // load the compiler options
        const rootDir = resolve(firstOf(opt.src) || this.getAppDir());
        const tsconfig = join(rootDir, 'tsconfig.json');
        const cmpOptions = _readCompilerOptions(tsconfig);
        // build the program
        const program = cmpOptions
            .then(cmdOpts => createProgram([join(rootDir, cmdOpts.baseUrl || '', 'main.ts')], cmdOpts));
        // the classes
        const classes = program.then(_getClasses);
        // process
        return from(classes.then(() => ''));
    }
}
