/* Copyright IBM Corp. 2017 */
import { join } from 'path';
import { merge, Observable } from 'rxjs';
import { distinct } from 'rxjs/operators';

import {
    APP_MODULE_FILE,
    CMD_REGISTER,
    LAYOUTS_FILE,
    OPTION_MODULE,
    OPTION_SKIP_REGISTRATION,
    OPTION_SOURCE,
    OPTION_VERBOSE,
} from '../constants';
import { firstOf } from '../utils/file.utils';
import { Options, optNormalizeDir } from './../options';
import { AbstractCommand } from './abstract.command';
import { Command } from './command';
import { createComponentOverview, createComponentTemplates, rewriteAppModule } from './create.components.command';
import { HELP_COMPONENT_SOURCE, HELP_MODULE_FILE, HELP_SKIP_REGISTRATION, HELP_VERBOSE } from './help';

/**
 * Generates angular components based on type and layout mapping definitions
 *
 * register layoutcomponents --src SOURCE_CODE_FOLDER [--module MODULENAME] [--verbose]
 */
export class RegisterLayoutComponentsCommand extends AbstractCommand implements Command {

    static readonly TYPE = 'layoutcomponents';

    static readonly COMMAND = CMD_REGISTER;

    constructor(private options: Options) {
        super(options);
    }

    printHelp(): void {
        // log the command line
        console.log(RegisterLayoutComponentsCommand.COMMAND, RegisterLayoutComponentsCommand.TYPE,
            `--${OPTION_SOURCE} <SRC>`,
            `[--${OPTION_MODULE} <MODULE_FILE>]`,
            `[--${OPTION_SKIP_REGISTRATION}]`,
            `[--${OPTION_VERBOSE}`
        );
        // log the parameters
        console.log(`--${OPTION_SOURCE} <SRC>`, ':', HELP_COMPONENT_SOURCE);
        console.log(`--${OPTION_MODULE} <MODULE_FILE>`, ':', HELP_MODULE_FILE);
        console.log(`--${OPTION_SKIP_REGISTRATION}`, ':', HELP_SKIP_REGISTRATION);
        console.log(`--${OPTION_VERBOSE}`, ':', HELP_VERBOSE);
    }

    process(): Observable<string> {
        // current options
        const opts = this.options;

        const srcDirOpt = optNormalizeDir(firstOf(opts.src) || join(this.getAppDir(), 'app'));
        const rxSrcDir = this.rxMkdirp(srcDirOpt);

        // module file
        const moduleFile = opts.module || join(srcDirOpt, APP_MODULE_FILE);
        const layoutsFile = join(srcDirOpt, LAYOUTS_FILE);

        // logger
        const logger = this.getLogger();

        // the compiler template
        const templates = createComponentTemplates(opts, logger);

        // log the name
        logger.next(`Source dir [${srcDirOpt}], ngModule [${moduleFile}], layouts [${layoutsFile}].`);

        // constructs the overview files
        const overview = createComponentOverview(rxSrcDir, templates, opts, logger);

        // registers the overview with the application
        const rewrite = rewriteAppModule(layoutsFile, moduleFile, opts, logger);

        // merge
        return merge(overview, rewrite).pipe(
            distinct()
        );
    }
}
