"use strict";
exports.__esModule = true;
var lodash_1 = require("lodash");
var path_1 = require("path");
var operators_1 = require("rxjs/operators");
var transliteration_1 = require("./../utils/transliteration");
var rx_dir_1 = require("./rx.dir");
function _getStyleExtension(aOptions) {
    return aOptions.scss ? '.scss' : '.css';
}
function _camelCase(aValue) {
    return transliteration_1.camelCase(aValue);
}
function _kebabCase(aValue) {
    return transliteration_1.kebabCase(aValue);
}
var COMPONENT_SUFFIX = 'Component';
var RENDERING_CONTEXT_SUFFIX = 'RenderingContext';
var COMPONENTS_PATH = 'components';
/**
 * Returns the directory name for the type
 *
 * @param aType     the type
 * @return the directory name
 */
function _getPathForType(aType, aOptions) {
    // access the type properties
    var name = aType.name;
    var tags = aType.tags || [];
    // convert tags into segments
    var segments = aOptions.flat ? [] : tags.map(_kebabCase).sort();
    segments.push(_kebabCase(name));
    // return this path
    return segments.join(path_1.sep);
}
function _createAbstractComponentName(aTypeName) {
    var name = lodash_1.upperFirst(_camelCase('abstract ' + aTypeName));
    return name.endsWith(COMPONENT_SUFFIX) ? name : name + COMPONENT_SUFFIX;
}
function _createTypeInterfaceName(aTypeName) {
    var name = lodash_1.upperFirst(_camelCase(aTypeName));
    return name.endsWith(RENDERING_CONTEXT_SUFFIX) ? name : name + RENDERING_CONTEXT_SUFFIX;
}
function _createTypeComponentName(aTypeName) {
    var name = lodash_1.upperFirst(_camelCase('type ' + aTypeName));
    return name.endsWith(COMPONENT_SUFFIX) ? name : name + COMPONENT_SUFFIX;
}
function _getTypeComponentPath(aType, aOptions) {
    // prepend by component
    return path_1.join(COMPONENTS_PATH, _getPathForType(aType, aOptions));
}
/**
 * Finds the directory tha the component should be created in. Tests
 * each potential source directory for an existing component file. If non exists, uses the
 * first source directory
 *
 * @param aType        type
 * @param aSources     sources directory
 *
 * @return the selected source directory
 */
function _getComponentsDir(aType, aSources, aOptions, aMkdirp, aLogger) {
    // check for the path suffix
    var typePath = _getTypeComponentPath(aType, aOptions);
    var className = _createAbstractComponentName(aType.name);
    var relPath = path_1.join(typePath, _camelCase(className) + '.ts');
    // map each source to a check
    return rx_dir_1.rxFindDir(aSources, relPath, aMkdirp).pipe(operators_1.tap(function (dir) { return aLogger.next("Selected component source directory [" + dir + "] for component [" + className + "]."); }));
}
/**
 * Constructs the relevant information for a components class
 *
 * @param aDir      base directory
 * @param aType     type object
 * @param aOptions  options
 * @param aLogger   logger
 *
 * @return the component object
 */
function _initComponentClass(aDir, aTypeDef, aOptions, aLogger) {
    // type object
    var type = aTypeDef.type;
    // check for the path suffix
    var typePath = _getTypeComponentPath(type, aOptions);
    var typeName = type.name;
    // component names
    var abstractComponentClass = _createAbstractComponentName(typeName);
    var typeComponentClass = _createTypeComponentName(typeName);
    var typeInterfaceClass = _createTypeInterfaceName(typeName);
    // references
    var abstractComponentRef = _camelCase(abstractComponentClass);
    var typeComponentRef = _camelCase(typeComponentClass);
    var typeInterfaceRef = _camelCase(typeInterfaceClass);
    // file names
    var abstractComponentFile = abstractComponentRef + '.ts';
    var typeComponentFile = typeComponentRef + '.ts';
    var typeInterfaceFile = typeInterfaceRef + '.ts';
    // template and style
    var typeTemplateFile = typeComponentRef + '.html';
    var typeStyleFile = typeComponentRef + _getStyleExtension(aOptions);
    // selector
    var typeSelector = _kebabCase('app ' + typeComponentClass);
    // the base path
    var folder = path_1.normalize(path_1.join(aDir, typePath));
    var result = {
        typeDef: aTypeDef, folder: folder, typeSelector: typeSelector, typeTemplateFile: typeTemplateFile, typeStyleFile: typeStyleFile, abstractComponentRef: abstractComponentRef, typeComponentRef: typeComponentRef, typeInterfaceRef: typeInterfaceRef, abstractComponentClass: abstractComponentClass, typeComponentClass: typeComponentClass, typeInterfaceClass: typeInterfaceClass, abstractComponentFile: abstractComponentFile, typeComponentFile: typeComponentFile, typeInterfaceFile: typeInterfaceFile
    };
    // log this
    aLogger.next("Type: " + typeName + ", ComponentClass: " + JSON.stringify(typeComponentClass) + ", Folder: " + folder);
    // ok
    return result;
}
function _createComponentClass(aTypeDef, aSources, aOptions, aMkdirp, aLogger) {
    // locate the component directory
    return _getComponentsDir(aTypeDef.type, aSources, aOptions, aMkdirp, aLogger).pipe(operators_1.first(), operators_1.map(function (dir) { return _initComponentClass(dir, aTypeDef, aOptions, aLogger); }), operators_1.mergeMap(function (clz) { return aMkdirp(clz.folder).pipe(operators_1.mapTo(clz)); }), operators_1.shareReplay());
}
var ComponentRegistry = /** @class */ (function () {
    function ComponentRegistry(aBaseFolders, aOptions, aMkdirp, aTypeRegistry, aLogger) {
        this.aBaseFolders = aBaseFolders;
        this.aOptions = aOptions;
        this.aMkdirp = aMkdirp;
        this.aTypeRegistry = aTypeRegistry;
        this.aLogger = aLogger;
        this.componentRegistry = {};
    }
    ComponentRegistry.prototype.registerType = function (aType) {
        return this.aTypeRegistry.registerType(aType);
    };
    ComponentRegistry.prototype.findTypeClass = function (aType) {
        return this.aTypeRegistry.findTypeClass(aType);
    };
    ComponentRegistry.prototype.findComponentClass = function (aType) {
        var _this = this;
        // the name
        var typeId = aType.id;
        // check if we are already searching
        var rxRunning = this.componentRegistry[typeId];
        if (!rxRunning) {
            // find the type
            var rxType = this.aTypeRegistry.findTypeClass(aType);
            // try to find the folder
            rxRunning = rxType.pipe(operators_1.mergeMap(function (type) { return _this.componentRegistry[typeId] = _createComponentClass(type, _this.aBaseFolders, _this.aOptions, _this.aMkdirp, _this.aLogger); }));
        }
        // ok
        return rxRunning;
    };
    return ComponentRegistry;
}());
exports.ComponentRegistry = ComponentRegistry;
